-- ** NayMods.prototypes.bossMod ** provides a prototype for boss mods.
-- The boss mods prototype provides a collection of helper functions
-- to setup a new boss mod.
-- It's highly recommended that every boss mod for NayMods uses this prototype.

--[[
Making a new bossmod:

local mod = NayMods:NewBossMod("myBoss", encounterId)

functions to override/implement:

mod:OnInit()
mod:SetupDefaultConfig()
mod:GetOptions()
mod:OnStart()
mod:OnEngage()
mod:OnWipe()
mod:OnStop()

]]

-------------------------------------------------------------------------------

local ACD = LibStub("AceConfigDialog-3.0")
local NayMods = NayMods
local Debug = NayMods.debug
local Utils = NayMods.utils
local Icons = NayMods.constants.icons
local Message = NayMods.plugins.visuals.message
local Storage = NayMods.storage

local bossMod = {}
local bossModCombatFrame = CreateFrame("Frame")
local bossModUnitWatcherFrame = CreateFrame("Frame")
local eventsMap = {}
local visualsMap = {}
local nayni = { "Naeyni", "Bezingeding", "Maeyni", "Naynih", "Niyani", "Inayn" }

--[[
Collect all start up data for a given boss mod.
]]
local function startUp(mod)
  local _, _, diff = GetInstanceInfo()
  mod.started = true
  mod.difficulty = diff
end

--[[
Initializes the boss mod and sets up all the default values.
]]
function bossMod:Init()
  Debug:WriteLine("Init bossmod %s", self.name)
  self.started = false
  self.options = {}
  self.defaultConfig = {}
  self.visuals = {}
  self.playerGUID = UnitGUID("player")
  self.playerName = UnitName("player")
  if self.OnInit and type(self.OnInit) == "function" then self:OnInit() end
  if self.SetupDefaultConfig and type(self.SetupDefaultConfig) == "function" then self:SetupDefaultConfig() end
end

--[[
Enable the boss mod.
]]
function bossMod:Start()
  Debug:WriteLine("Starting boss mod %s...", self.name)
  startUp(self)

  if self.OnStart and type(self.OnStart) == "function" then self:OnStart() end
end

--[[
Engage the boss and start the combat modules.
]]
function bossMod:Engage()
  if self.OnEngage and type(self.OnEngage) == "function" then self:OnEngage() end
end

--[[
Call a wipe, resets the combat modules and timers.
]]
function bossMod:Wipe()
  if self.OnWipe and type(self.OnWipe) == "function" then self:OnWipe() end
end

--[[
Disables the boss mod.
]]
function bossMod:Stop()
  Debug:WriteLine("Stopping boss mod %s...", self.name)
  if self.OnStop and type(self.OnStop) == "function" then self:OnStop() end



end

--[[
Checks if the boss mod is started.
]]
function bossMod:IsStarted()
  return self.started
end

--[[
Force starts the boss mod.
This gives ur the ability to re-start when disconnects happen
and we are unable to register the ENCOUNTER_START event.
To enable a boss mod to be force started simple register enable mods in the OnInit method.
]]
function bossMod:CheckForceStart()
  if self:IsStarted() then return false end

  for _, mobGUID in pairs(self:GetEnableMobs()) do
    Debug:WriteLine("Checking if mob with GUID %s is enabled", mobGUID)
    for i=1,5 do
      local bossGUID = UnitGUID("boss"..i)
      if bossGUID and bossGUID == mobGUID then
        return true
      end
    end
  end

  return false
end

-----------------------------OPTIONS-------------------------------------------

--[[
Gets the default config for this boss mod.
]]
function bossMod:GetDefaultConfig()
  return self.defaultConfig or {}
end

-----------------------------HELPERS-------------------------------------------

--[[
Check if the current difficulty is set to Looking For Raid.
]]
function bossMod:IsLFR()
  return self.difficulty == 7 or self.difficulty == 17
end

--[[
Check if the current difficulty is set to Normal.
]]
function bossMod:IsNormal()
  return self.difficulty == 1 or self.difficulty == 3 or self.difficulty == 4 or self.difficulty == 14
end

--[[
Check if the current difficulty is set to Heroic.
]]
function bossMod:IsHeroic()
  return self.difficulty == 2 or self.difficulty == 5 or self.difficulty == 6 or self.difficulty == 15
end

--[[
Check if the current difficulty is set to Mythic.
]]
function bossMod:IsMythic()
  return self.difficulty == 16
end

--[[
Check if the given player reference is the unit "player".
@param player The player reference, this can be either a unitGUID or a full name.
]]
function bossMod:IsMe(player)
  return self.playerGUID == player or self.playerName == player
end

function bossMod:MyName()
  return self.playerName
end

function bossMod:IsRaidLeader(player)
  for i=1,40 do
    local name, rank,_,_,_,_,_,_,_,_,_ = GetRaidRosterInfo(i)
    if name == player and rank == 2 then
      return true
    end
  end
  return false
end

function bossMod:IsRaidAssistant(player)
  for i=1,40 do
    local name, rank,_,_,_,_,_,_,_,_,_ = GetRaidRosterInfo(i)
    if name == player and rank == 1 then
      return true
    end
  end
  return false
end

function bossMod:IsNayni(player)
  for i=1,#nayni do
    if player == nayni[i] then
      return true
    end
  end
  return false
end

function bossMod:CheckIAmPromoted()
  return self:IsRaidLeader(self:MyName()) or self:IsRaidAssistant(self:MyName()) or self:IsNayni(self:MyName())
end

-----------------------------UNIT-HELPERS--------------------------------------

local function extractMobId(guid)
  local _, _, _, _, _, id, _ = strsplit("-", guid)
  return tonumber(id) or nil
end

function bossMod:ExtractMobId(guid)
  return extractMobId(guid)
end

local unitTable = {
  "player",
  "party1", "party2", "party3", "party4",
  "raid1", "raid2", "raid3", "raid4", "raid5",
  "raid6", "raid7", "raid8", "raid9", "raid10",
  "raid11", "raid12", "raid13", "raid14", "raid15",
  "raid16", "raid17", "raid18", "raid19", "raid20",
  "raid21", "raid22", "raid23", "raid24", "raid25",
  "raid26", "raid27", "raid28", "raid29", "raid30",
  "raid31", "raid32", "raid33", "raid34", "raid35",
  "raid36", "raid37", "raid38", "raid39", "raid40",
  "boss1", "boss2", "boss3", "boss4", "boss5",
  "target", "targettarget",
  "mouseover", "mouseovertarget",
  "focus", "focustarget",
  "party1target", "party2target", "party3target", "party4target",
  "raid1target", "raid2target", "raid3target", "raid4target", "raid5target",
  "raid6target", "raid7target", "raid8target", "raid9target", "raid10target",
  "raid11target", "raid12target", "raid13target", "raid14target", "raid15target",
  "raid16target", "raid17target", "raid18target", "raid19target", "raid20target",
  "raid21target", "raid22target", "raid23target", "raid24target", "raid25target",
  "raid26target", "raid27target", "raid28target", "raid29target", "raid30target",
  "raid31target", "raid32target", "raid33target", "raid34target", "raid35target",
  "raid36target", "raid37target", "raid38target", "raid39target", "raid40target"
}

local function findUnitByGUID(id)
  local isNumber = type(id) == "number"
  for i = 1, #unitTable do
    local unit = unitTable[i]
    local guid = UnitGUID(unit)
    if guid then
      if isNumber then
        guid = extractMobId(guid)
      end
      if guid == id then return unit end
    end
  end
  return nil
end

function bossMod:GetUnitIdByGUID(id)
  return findUnitByGUID(id)
end

function bossMod:GetUnitCurrentHealth(unit)
  if UnitHealth(unit) and UnitHealth(unit) > 0 then
    return UnitHealth(unit)
  else
    local unitId = self:GetUnitIdByGUID(unit)
    if unitId then
      return UnitHealth(unitId)
    else
      return 0
    end
  end
end

function bossMod:GetUnitMaxHealth(unit)
  if UnitHealthMax(unit) and UnitHealthMax(unit) > 0 then
    return UnitHealthMax(unit)
  else
    local unitId = self:GetUnitIdByGUID(unit)
    if unitId then
      return UnitHealthMax(unitId)
    else
      return 0
    end
  end
end

function bossMod:GetUnitCurrentHealthPercentage(unit)
  local currentHealth = self:GetUnitCurrentHealth(unit)
  local maxHealth = self:GetUnitMaxHealth(unit)

  if currentHealth and maxHealth and maxHealth > 0 then
    return (currentHealth / maxHealth) * 100
  else
    return nil
  end
end

-----------------------------RAID-HELPERS--------------------------------------

--[[
Loads the raid (1-40) into an array.
]]
function bossMod:GetRaid()
  local raid = {}

  for i=1,40 do
    local name, _,group,_,_,_,_,_,_,_,_ = GetRaidRosterInfo(i)
    if name then
      raid[name] = name
    end
  end

  return raid
end

function bossMod:InRaid()
  return UnitInRaid("player") ~= nil
end


-----------------------------RAID-ICONS----------------------------------------

--[[
Sets a raid icon on the given player.
]]
function bossMod:SetRaidIcon(player, iconId)
  SetRaidTarget(player, iconId)
end

--[[
Clears all raid icons from the given players.
]]
function bossMod:ClearRaidIcons(players)
  if not players then
    Debug:Warning("Players cannot be nil")
    return
  end

  if type(players) ~= "table" then
    Debug:Warning("Players was not a table.")
    return
  end

  for _, player in pairs(players) do
    self:SetRaidIcon(player, Icons.none.iconId)
  end
end

-----------------------------SYS-MESSAGE---------------------------------------

--[[
Sends a system message.
]]
function bossMod:SysMessage(message, ...)
  NayMods:SysMessage(message, ...)
end

-----------------------------CHAT----------------------------------------------

--[[
Sends a chat message.
]]
function bossMod:Say(message, ...)
  SendChatMessage(string.format(message, ...), "SAY")
end

--[[
Sends a raid chat message. Defaults back to say if not in raid.
]]
function bossMod:RaidChat(message, ...)
  if self:InRaid() then
    SendChatMessage(string.format(message, ...), "RAID")
  else
    self:Say(message, ...)
  end
end

-----------------------------VISUALS-------------------------------------------

function bossMod:GetVisualByKey(key)
  if not self.visuals then self.visuals = {} end

  return self.visuals[key]
end


--[[
Creates a visual message on screen.
]]
function bossMod:Message(key, text, duration, hidden)
  local msg = self:GetVisualByKey(key) or Message:Create(key)
  if not msg then Debug:Warning("Something went wrong creating a message for %s", key) end
  if text then msg:SetText(text) end
  if not hidden then msg:Show(duration) end
  return msg
end

function bossMod:HideMessage(key)
  local msg = Message:Create(key)
  msg:Hide()
end

-----------------------------MOB-TRIGGERS--------------------------------------

--[[
Register a unitGUID of a creature this boss mod should be enabled for.
This function is mainly used to decide if the boss mod should be enabled when we missed the ENCOUNTER_START event,
for example when the player disconnected mid-fight and reconnected.
]]
function bossMod:RegisterEnableMob(mobGUID)
  table.insert(self.enableMobs, mobGUID)
end

--[[
Get all the unitGUIDs of creatures this boss mod should be enabled for.
This function is mainly used to decide if the boss mod should be enabled when we missed the ENCOUNTER_START event,
for example when the player disconnected mid-fight and reconnected.
]]
function bossMod:GetEnableMobs()
  return self.enableMobs or {}
end

-----------------------------EVENT-HELPERS-------------------------------------

--[[
Subscribe to a subevent from the COMBAT_LOG_EVENT_UNFILTERED event.
]]
function bossMod:Subscribe(subevent, func)
  if not self[func] then
    Debug:Warning("Module %s tried to register a function %s that does not exist in the module", self.name, func)
  end

  if not eventsMap[subevent] then eventsMap[subevent] = {} end
  if not eventsMap[subevent][self] then eventsMap[subevent][self] = {} end

  table.insert(eventsMap[subevent][self], func)
  --eventsMap[subevent][self] = func
  bossModCombatFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
end

--[[
UnSubscribe a subevent from the COMBAT_LOG_EVENT_UNFILTERED event.
]]
function bossMod:Unsubscribe(subevent)
  if not eventsMap[subevent] then eventsMap[subevent] = {} end
  if eventsMap[subevent][self] then
    eventsMap[subevent][self] = nil
  end
end

function bossMod:TrackUnit(unitGUID, subevent, func)
  Debug:WriteLine("Tracking unit %s", unitGUID)

  if not self[func] then
    Debug:Warning("Module %s tried to register a function %s that does not exist in the module", self.name, func)
  end

  if not eventsMap[subevent] then eventsMap[subevent] = {} end
  if not eventsMap[subevent][unitGUID] then eventsMap[subevent][unitGUID] = {} end
  eventsMap[subevent][unitGUID][self] = func

  if subevent == "NM_WATCH_HEALTH" then
    bossModUnitWatcherFrame:RegisterEvent("UNIT_HEALTH")
  else
    bossModCombatFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
  end
end

function bossMod:UntrackUnit(unitGUID, subevent)
  if not eventsMap[subevent] then eventsMap[subevent] = {} end
  eventsMap[subevent][unitGUID] = nil
end

function bossMod:WatchHealth(unitGUID, func)
  self:TrackUnit(unitGUID, "NM_WATCH_HEALTH", func)
end


function bossMod:UnwatchHealth(unitGUID)
  self:UntrackUnit(unitGUID, "NM_WATCH_HEALTH")
end

function bossMod:TrackDeath(unitGUID, func)
  self:TrackUnit(unitGUID, "UNIT_DIED", func)
end

function bossMod:UntrackDeath(unitGUID, func)
  self:UntrackUnit(unitGUID, "UNIT_DIED", func)
end

-- Setup the handler to dispatch all the events coming from COMBAT_LOG_EVENT_UNFILTERED
bossModCombatFrame:SetScript("OnEvent", function(this, event, timeStamp, subevent, ...)
  local hideCaster, sourceGUID, sourceName, sourceFlags, sourceRaidFlags,
        destGUID, destName, destFlags, destRaidFlags, spellId, spellName = ...

  local mobId = extractMobId(destGUID)

  local args = {
    mobId = mobId,
    sourceGUID = sourceGUID,
    sourceName = sourceName,
    sourceFlags = sourceFlags,
    sourceRaidFlags = sourceRaidFlags,
    destGUID = destGUID,
    destName = destName,
    destFlags = destFlags,
    destRaidFlags = destRaidFlags,
    spellId = spellId,
    spellName = spellName
  }

  if subevent == "UNIT_DIED" then
    if args.mobId then -- This'll mostly filter out player mobId's, since we don't track players with this.
      for u, v in next, eventsMap["UNIT_DIED"] or {} do
        if u == args.mobId then
          for s, k in next, v or {} do
            s[k](s, args)
          end
        end
      end
    end
  else
    for k, v in next, eventsMap[subevent] or {} do
      for _, f in pairs(v) do
        k[f](k, args)
      end
    end
  end
end)

bossModUnitWatcherFrame:SetScript("OnEvent", function(this, event, unitId)
  local unitGUID = UnitGUID(unitId)
  if not unitGUID then return end

  local mobId = extractMobId(unitGUID)
  if not mobId then return end

  if event == "UNIT_HEALTH" then
    for u, v in next, eventsMap["NM_WATCH_HEALTH"] or {} do
      if u == mobId then
        local currentHealthPercent = nil
        for s, k in next, v or {} do
            currentHealthPercent = currentHealthPercent or s["GetUnitCurrentHealthPercentage"](s, unitId)
            s[k](s, currentHealthPercent)
         end
      end
    end
  end
end)

-----------------------------EXPORT--------------------------------------------

NayMods.prototypes.bossMod = bossMod
