-- ** NayMods.Core ** holds the core module of NayMods.
-------------------------------------------------------------------------------

local NayMods = NayMods
local Debug = NayMods.debug
local Options = NayMods.options
local Storage = NayMods.storage
local Utils = NayMods.utils
local Transmission = NayMods.transmission
local Timer = NayMods.timer

local eventsMap = {}
local nayModsCombatFrame = CreateFrame("Frame")
local myName = UnitName("player")

local table_sort_desc = Utils.TableSortDescending

local VERSION = tonumber(GetAddOnMetadata("NayMods", "version"))
local RECEIVED_VERSIONS = {}
local VERSIONS_OVERVIEW = {}
local HIGHER_VERSION_RECEIVED = false

local next = next

-----------------------------INITIALIZE----------------------------------------

--[[
Gets/Creates the default config options for NayMods.
]]
function NayMods:GetDefaultConfig()
  local defaults = {
    char = {
      visuals = {
        message = {
          enabled = true,
          height = 200,
          width = 200,
          xOffset = 0,
          yOffset = 0,
          point = "CENTER",
          relativePoint = "CENTER",
          font = "Friz Quadrata TT",
          fontSize = 24,
          textAlign = "CENTER"
        }
      }
    }
  }

  for _, bossMod in self:GetBossMods() do
    bossMod:SetupDefaultConfig()
    defaults = Utils:TableMerge(defaults, bossMod:GetDefaultConfig())
  end

  return defaults
end

--[[
OnInitialize handler for NayMods.
]]
function NayMods:OnInitialize()
  Storage:Initialize()
  Storage:LoadDefaultConfig(self:GetDefaultConfig())

  for _, bossMod in self:GetBossMods() do
    bossMod:Init()
  end

  Options:Initialize()
  Debug:WriteLine("NayMods Initialize completed!")
end

--[[
OnEnable handler for NayMods.
]]
function NayMods:OnEnable()
  Timer.After(5, function()
    self:SysMessage("Welcome to NayMods version %.1f. Type /nm to access in-game options.", VERSION)
  end)
  self:ForceStartBossMods()
end

-----------------------------HELPERS-------------------------------------------

--[[
Sends a system message to the chat frame.
]]
function NayMods:SysMessage(message, ...)
  local sysColor = "E6E600"
  local nayModsColor = "9999FF"

  print(Utils:Color("NayMods: ", nayModsColor) .. Utils:Color(string.format(message, ...), sysColor))
end

function NayMods:GetRaid()
  local raid = {}

  for i=1,40 do
    local name, _,group,_,_,_,_,_,_,_,_ = GetRaidRosterInfo(i)
    if name then
      raid[name] = name
    end
  end

  return raid
end

-----------------------------VERSION-------------------------------------------

--[[
Compare this version of NayMods with all the received versions made during a version check.
]]
function NayMods:CompareVersions()
  if RECEIVED_VERSIONS ~= nil and not Utils:TableEmpty(RECEIVED_VERSIONS) then
    table_sort_desc(RECEIVED_VERSIONS)
    Debug:Dump(RECEIVED_VERSIONS)
    local highestVersion = RECEIVED_VERSIONS[1]
    Debug:WriteLine("Highest version received = %s", highestVersion)
    Debug:WriteLine("My version = %s", VERSION)
    if highestVersion > VERSION and not HIGHER_VERSION_RECEIVED then
      HIGHER_VERSION_RECEIVED = true
      self:SysMessage("Your version of NayMods seems to be out of date. Consider updating to the newest version!")
    end
  end

  RECEIVED_VERSIONS = {}
end

--[[
Initiate a version check for NayMods.
]]
function NayMods:VersionCheck()
  Debug:WriteLine("Doing version check!")
  Transmission:Send("VERSION_CHCK", { version = VERSION }, "RAID")
  Timer.After(10, function(...) NayMods:CompareVersions() end)
end

--[[
Callback when we receive a version.
]]
function NayMods:onVersionReceived(key, data, dist, sender)
  Debug:WriteLine("Received Version from %s", sender)
  if sender == myName then return end

  local versionString = data["version"] or "0.0"
  local versionNumber = tonumber(versionString) or 0.0
  table.insert(RECEIVED_VERSIONS, versionNumber)
end


function NayMods:RequestVersionsOverview()
  Debug:WriteLine("Requesting version overview!")
  self:SysMessage("Performing raid-wide version check, please wait...")
  Transmission:Send("VERSION_OVERVIEW_REQUEST", {}, "RAID")

  local versionNumber = tonumber(VERSION) or 0.0
  if not VERSIONS_OVERVIEW then VERSIONS_OVERVIEW = {} end
  if not VERSIONS_OVERVIEW[versionNumber] then VERSIONS_OVERVIEW[versionNumber] = {} end
  table.insert(VERSIONS_OVERVIEW[versionNumber], Utils:ColorByClass(myName))

  Timer.After(2, function(...) NayMods:ShowVersionsOverview() end)
end

function NayMods:OnRequestVersionsOverview(key, data, dist, sender)
  Debug:WriteLine("Received a version overview request from %s!", sender)
  Transmission:Send("VERSION_OVERVIEW_RESPONSE", { version = VERSION }, "WHISPER", sender)
end

function NayMods:OnVersionOverviewResponse(key, data, dist, sender)
  if sender == myName then return end

  local versionString = data["version"] or "0.0"
  local versionNumber = tonumber(versionString) or 0.0

  if not VERSIONS_OVERVIEW then VERSIONS_OVERVIEW = {} end
  if not VERSIONS_OVERVIEW[versionNumber] then VERSIONS_OVERVIEW[versionNumber] = {} end

  table.insert(VERSIONS_OVERVIEW[versionNumber], Utils:ColorByClass(sender))
end

function NayMods:ShowVersionsOverview()
  local nayModsColor = "9999FF"
  local sysColor = "E6E600"
  local raid = self:GetRaid()
  local hasNayMods = {}

  self:SysMessage("==-----VERSION OVERVIEW-----==")

  if VERSIONS_OVERVIEW ~= nil and not Utils:TableEmpty(VERSIONS_OVERVIEW) then
    table.sort(VERSIONS_OVERVIEW, table_sort_desc)
    for i=table.maxn(VERSIONS_OVERVIEW), 0, -0.1 do
      local index = tonumber(string.format("%.1f", i))
      if VERSIONS_OVERVIEW[index] then
        print(Utils:Color("NayMods: ", nayModsColor) .. Utils:Color(string.format("# V.%.1f: ", index), sysColor))
        print(Utils:Color("NayMods: ", nayModsColor) .. table.concat(VERSIONS_OVERVIEW[index], ", "))

        for _, n in pairs(VERSIONS_OVERVIEW[index]) do
          hasNayMods[n] = n
        end
      end
    end

    local noNayMods = {}

    for k, v in pairs(raid) do
      if not hasNayMods[Utils:ColorByClass(v)] then
        table.insert(noNayMods, Utils:ColorByClass(v))
      end
    end

    if noNayMods ~= nil and not Utils:TableEmpty(noNayMods) then
      print(Utils:Color("NayMods: ", nayModsColor) .. Utils:Color("# Not installed: ", sysColor))
      print(Utils:Color("NayMods: ", nayModsColor) .. table.concat(noNayMods, ", "))
    end
  end

  self:SysMessage("==---------------------------------==")

  VERSIONS_OVERVIEW = {}
end

-----------------------------BOSS-MODS-----------------------------------------

--[[
Creates a new boss mod module for NayMods.
]]
function NayMods:NewBossMod(name, encounterId)
  local bossMod = self:NewModule(name, NayMods.prototypes.bossMod)
  bossMod:SetDefaultModuleState(false)

  bossMod.name = name
  bossMod.encounterId = encounterId
  bossMod.enableMobs = {}

  return bossMod
end

--[[
Get all the registered boss mods in NayMods.
]]
function NayMods:GetBossMods()
  return self:IterateModules()
end

function NayMods:ForceStartBossMods()
  for _, bossMod in self:GetBossMods() do
    Debug:WriteLine("Checking if bossmod %s should be started", bossMod.name)
    if bossMod:CheckForceStart() then
      bossMod:Start()
      bossMod:Engage()
    end
  end
end

-----------------------------ENCOUNTER-HOOKS-----------------------------------

--[[
Handler for the ECOUNTER_START event.
@param encounterId The Id of the encounter that just started.
]]
function NayMods:ENCOUNTER_START(encounterId)
  Debug:WriteLine("Encounter started id = %s", encounterId)
  for _, bossMod in self:GetBossMods() do
    Debug:WriteLine("Module %s - encounterId = %s", bossMod.name, bossMod.encounterId)
    if bossMod.encounterId == encounterId then
      bossMod:Start()
      bossMod:Engage()
    end
  end
end

--[[
Handler for the ECOUNTER_END event.
@param encounterId The Id of the encounter that just ended.
]]
function NayMods:ENCOUNTER_END(encounterId)
  Debug:WriteLine("Encounter ended id = %s", encounterId)
  for _, bossMod in self:GetBossMods() do
    if bossMod.encounterId == encounterId then
      bossMod:Wipe()
      bossMod:Stop()
    end
  end
end

-----------------------------EVENT-REGISTRATION--------------------------------

--[[
Register to an event.
]]
function NayMods:RegisterEvent(event, func)
  Debug:WriteLine("Registering %s with %s as callback", event, func)
  if not self[func] then
    Debug:Warning("Tried to register a function %s that does not exist", func)
  end
  if not eventsMap[event] then eventsMap[event] = {} end
  if not eventsMap[event][self] then eventsMap[event][self] = {} end

  table.insert(eventsMap[event][self], func)

  nayModsCombatFrame:RegisterEvent(event)
end

-- Handler for all the registered events in NayMods.
nayModsCombatFrame:SetScript("OnEvent", function(this, event, ...)
  for k, v in next, eventsMap[event] do
    for _, f in pairs(v) do
      k[f](k, ...)
    end
  end
end)

--[[
General callback when the group roster updates.
Mostly need this to send refreshed data to the options dialog.
]]
function NayMods:GROUP_ROSTER_UPDATE(...)
  Options:Refresh()
end

Transmission:Register("VERSION_CHCK", function(...) NayMods:onVersionReceived(...) end)
Transmission:Register("VERSION_OVERVIEW_REQUEST", function(...) NayMods:OnRequestVersionsOverview(...) end)
Transmission:Register("VERSION_OVERVIEW_RESPONSE", function(...) NayMods:OnVersionOverviewResponse(...) end)

NayMods:RegisterEvent("GROUP_ROSTER_UPDATE", "VersionCheck")
NayMods:RegisterEvent("GROUP_ROSTER_UPDATE", "GROUP_ROSTER_UPDATE")
NayMods:RegisterEvent("ENCOUNTER_START", "ENCOUNTER_START")
NayMods:RegisterEvent("ENCOUNTER_END", "ENCOUNTER_END")
