-- ** NayMods.utils ** provides general utilities used all over NayMods
-------------------------------------------------------------------------------

local NayMods = NayMods
local Utils = {}

-----------------------------PROTOTYPING---------------------------------------

--[[
Provides us with a general prototyping or class-feel system.
A module can be anything that's based off a base class or prototype object.
This function is mostly used to easely make plugins and extend from the already build prototypes.

Every module made with this function can be instantiated with the create function provided.
to hook into the object creation you're allowed to implement an initialize method. Any parameter
passed to the create method will be passed to the initliaze method aswell.

This function provides deep nested prototyping.
]]
function Utils:NewModule(baseClass)
  local new_class = {}
  local class_mt = { __index = new_class }

  function new_class:Create(...)
    local newinst = {}
    setmetatable(newinst, class_mt)
    newinst:initialize(...)
    return newinst
  end

  if baseClass then
    setmetatable(new_class, { __index = baseClass})
  end

  return new_class
end

-----------------------------TABLE-UTILS---------------------------------------

--[[
Provides the functionality of a deep copy of a table. This includes meta tables to make a true copy that is
not linked to the original in any way.
]]
function Utils:TableCopy(orig)
    local orig_type = type(orig)
    local copy
    if orig_type == 'table' then
        copy = {}
        for orig_key, orig_value in next, orig, nil do
            copy[self:TableCopy(orig_key)] = self:TableCopy(orig_value)
        end
        setmetatable(copy, self:TableCopy(getmetatable(orig)))
    else -- number, string, boolean, etc
        copy = orig
    end
    return copy
end

--[[
Provides the functionality to merge two given tables with eachother.
The first passed table will be merged with the second, the first table will be mutated to hold the merged value.
]]
function Utils:TableMerge(a, b)
  if type(a) == 'table' and type(b) == 'table' then
      for k,v in pairs(b) do if type(v)=='table' and type(a[k] or false)=='table' then self:TableMerge(a[k],v) else a[k]=v end end
  end
  return a
end

--[[
Checks if a given table is empty.
]]
function Utils:TableEmpty(tbl)
  if next(tbl) == nil then
    return true
  end
  return false
end

--[[
Checks if an array contains a given value.
]]
function Utils:ArrayContains(array, value)
  for _, v in pairs(array) do
    if v == value then
      return true
    end
  end
  return false
end

--[[
Sort a given table in a descending order.
]]
function Utils:TableSortDescending(tbl)
  if tbl ~= nil and type(tbl) == 'table' then
    table.sort(tbl, function(a, b) return a > b end)
  end
end

--[[
Sort a given table in an ascending order.
]]
function Utils:TableSortAscending(tbl)
  if tbl ~= nil and type(tbl) == 'table' then
    table.sort(tbl, function(a, b) return a < b end)
  end
end

-----------------------------COLOR-UTILS---------------------------------------

--[[
Colors the given message with the given color in hexadecimal format.
]]
function Utils:Color(message, colorHex)
  if message and colorHex then
    return "|cFF" .. colorHex .. "" .. message .. "|r"
  else
    return message
  end
end

--[[
Colors the given unit name by it's class color.
]]
function Utils:ColorByClass(unitName)
  if unitName then
    local classColor = "ffffff"
    local class, classFileName = UnitClass(unitName)

    if classFileName and RAID_CLASS_COLORS[classFileName] and RAID_CLASS_COLORS[classFileName]["colorStr"]  then
      classColor = RAID_CLASS_COLORS[classFileName].colorStr
    end
    return "|c" .. classColor .. "" .. unitName .. "|r"
  end
end

function Utils:Round(num, idp)
  local mult = 10^(idp or 0)
  return math.floor(num * mult + 0.5) / mult
end

-----------------------------CLASS-UTILS---------------------------------------

local classMap = {
  [1] = {
    name = "Warrior",
    canBeTank = true,
    canBeHealer = false,
    canBeMelee = true,
    canBeRanged = false,
  },
  [2] = {
    name = "Paladin",
    canBeTank = true,
    canBeHealer = true,
    canBeMelee = true,
    canBeRanged = false,
  },
  [3] = {
    name = "Hunter",
    canBeTank = false,
    canBeHealer = false,
    canBeMelee = false,
    canBeRanged = true,
  },
  [4] = {
    name = "Rogue",
    canBeTank = false,
    canBeHealer = false,
    canBeMelee = true,
    canBeRanged = false,
  },
  [5] = {
    name = "Priest",
    canBeTank = false,
    canBeHealer = true,
    canBeMelee = false,
    canBeRanged = true,
  },
  [6] = {
    name = "Deathknight",
    canBeTank = true,
    canBeHealer = false,
    canBeMelee = true,
    canBeRanged = false,
  },
  [7] = {
    name = "Shaman",
    canBeTank = false,
    canBeHealer = true,
    canBeMelee = true,
    canBeRanged = true,
  },
  [8] = {
    name = "Mage",
    canBeTank = false,
    canBeHealer = false,
    canBeMelee = false,
    canBeRanged = true,
  },
  [9] = {
    name = "Warlock",
    canBeTank = false,
    canBeHealer = false,
    canBeMelee = false,
    canBeRanged = true,
  },
  [10] = {
    name = "Monk",
    canBeTank = true,
    canBeHealer = true,
    canBeMelee = true,
    canBeRanged = false,
  },
  [11] = {
    name = "Druid",
    canBeTank = true,
    canBeHealer = true,
    canBeMelee = true,
    canBeRanged = true,
  },
}

--[[
Check if the given player is melee.
** NOTE ** This method avoids using GetSpecializationInfo or a similar methods as these often return nil values
because it requires us to send an inspect query first.
What happens here is that we use a simple UnitClass call to retrieve the class index and match it with our
existing class map.
A decision then gets made based on the players currents max power (mana).
@returns true if the player is most likely melee, false if not.
]]
function Utils:IsMelee(playerName)
  local _, _, classIndex = UnitClass(playerName)
  local maxPower = nil

  if not classIndex then
    return false
  end

  if not classMap[classIndex] then
    Debug:Warning("Could not match class index %s for player %s", classIndex, playerName)
    return false
  end

  local classInfo = classMap[classIndex]

  if classInfo.canBeRanged or classInfo.canBeHealer then
    maxPower = UnitManaMax(playerName)
  end

  if classInfo.canBeMelee then
    if maxPower == nil or maxPower < 33000 then
      return true
    end
  end

  return false
end

-----------------------------EXPORT--------------------------------------------

NayMods.utils = Utils
