-- ** NayMods.plugins.visuals.prototype **
-- holds a prototype for any visual on screen.
-------------------------------------------------------------------------------

local NayMods = NayMods
local Debug = NayMods.debug
local Utils = NayMods.utils
local Storage = NayMods.storage

local visualPrototype = Utils:NewModule(nil)

--[[
Hook into module creation.
]]
function visualPrototype:initialize(key)
  self.key = key
  self.frame = self:findOrCreateFrame(key, self:GetConfig())
  self.totalElapsed = 0

  if self.OnInit and type(self.OnInit) == "function" then
    self:OnInit(key, self:GetConfig())
  end
end

--[[
Gets this visuals current config.
]]
function visualPrototype:GetConfig()
  local defaultConfig = Utils:TableCopy(Storage.db.char.visuals.message)
  local config = Utils:TableMerge(defaultConfig, Storage.db.char.visuals[self.key] or nil)
  return config
end

--[[
Updates this visuals config.
]]
function visualPrototype:UpdateConfig(config)
  Storage.db.char.visuals[self.key] = Utils:TableMerge(Storage.db.char.visuals[self.key] or {}, config)
end

--[[
Finds of creates the frame associated with this visual.
]]
function visualPrototype:findOrCreateFrame(key, config)
  local frame = {}

  -- Find the frame in global space, if found re-use it, else create new one.
  if _G["NayMods_Visual_" .. key] then
    --Debug:WriteLine("Frame for visual with key %s already exists, reusing it.", key)
    frame = _G["NayMods_Visual_" .. key]
  else
    frame = CreateFrame("Frame", "NayMods_Visual_" .. key, UIParent)
    local movingTexture = frame:CreateTexture("NayMods_Visual_MovingTexture_" .. key, "ARTWORK")
    movingTexture:SetAllPoints()
    movingTexture:SetTexture(1.0, 0.5, 0)
    movingTexture:SetAlpha(0.5)
    movingTexture:Hide()
  end

  -- Basic config for the base frame.
  frame.fadetime = 0
  frame.wait = 0
  frame:SetFrameStrata("DIALOG")
  frame:SetHeight(config.height)
  frame:SetWidth(config.width)
  frame:SetPoint(config.point, UIParent, config.relativePoint, config.xOffset, config.yOffset)
  frame:Hide()

  -- Register the OnUpdate with a self callback.
  frame:SetScript("OnUpdate", function(...)
    local _, elapsed = ...

    if self.onUpdate and type(self.onUpdate) == "function" then
      self:onUpdate(elapsed)
    end
  end)

  return frame
end

--[[
Register an on update callback to this visual.
]]
function visualPrototype:RegisterOnUpdate(key, func)
  self.onUpdates = self.onUpdates or {}
  if self.onUpdates[key] then
    Debug:Warning("OnUpdate with key %s is already registered.", key)
    return
  end

  self.onUpdates[key] = func
  --table.insert(self.onUpdates, func)
end

function visualPrototype:UnregisterOnUpdate(key)
  self.onUpdates = self.onUpdates or {}

  if not self.onUpdates[key] then
    Debug:Warning("OnUpdate with key %s not found.", key)
    return
  end

  self.onUpdates[key] = nil
end

--[[
General onUpdate handler for this visual.
]]
function visualPrototype:onUpdate(elapsed)
  self.totalElapsed = self.totalElapsed + elapsed

  for _, f in pairs(self.onUpdates) do
    f(self, self.totalElapsed)
  end

  if not self.duration then
    return
  end

  if self.totalElapsed >= self.duration then
    self:Hide()
    self.totalElapsed = 0
  end
end

function visualPrototype:IsEnabled()
  return self:GetConfig().enabled or false
end

--[[
Show this visual, optionally give up a duration on how long you want it to be shown.
]]
function visualPrototype:Show(duration)
  if not self:IsEnabled() then return end

  self.duration = duration or nil

  if self.OnShow and type(self.OnShow) == "function" then
    self:OnShow()
  end

  self.frame:Show()
end

--[[
Hides this visual.
]]
function visualPrototype:Hide()
  if self.OnHide and type(self.OnHide) == "function" then
    self:OnHide()
  end
  self.frame:Hide()
end

--[[
Shows the anchors of this visual.
]]
function visualPrototype:showAnchor()
  _G["NayMods_Visual_MovingTexture_" .. self.key]:Show()
end

--[[
Hides the anchors of this visual.
]]
function visualPrototype:hideAnchor()
    _G["NayMods_Visual_MovingTexture_" .. self.key]:Hide()
end

--[[
Disable movement of this visual.
]]
function visualPrototype:disableMoving()
  local frame = self.frame
  self:hideAnchor()
  frame:SetMovable(false)
  frame:EnableMouse(false)
end

--[[
Enable movement of this visual.
]]
function visualPrototype:enableMoving()
  local frame = self.frame
  self:showAnchor()
  frame:SetMovable(true)
  frame:EnableMouse(true)
  frame:RegisterForDrag("LeftButton")
  frame:SetScript("OnDragStart", frame.StartMoving)
  frame:SetScript("OnDragStop", function(...)
    frame:StopMovingOrSizing()
    local point, relativeTo, relativePoint, xOffset, yOffset = frame:GetPoint(0)
    self:UpdateConfig({
      point = point,
      relativePoint = relativePoint,
      xOffset = xOffset,
      yOffset = yOffset
    })
  end)
end

--[[
Checks if this visual is currently movable by the user.
]]
function visualPrototype:IsMovable()
  return self.frame:IsMovable()
end

--[[
Toggle the moving anchors on/off.
]]
function visualPrototype:ToggleAnchor()
  if not self:IsEnabled() then
    NayMods:SysMessage("TIP!: Enable the visual before toggling the anchor.")
  end

  if self:IsMovable() then
    self:disableMoving()
    self:Hide()
  else
    self:enableMoving()
    self:Show()
  end
end

-----------------------------ACE3-OPTIONS-UTILS--------------------------------

function visualPrototype:EnableOption(order)
  return {
    type = "toggle",
    name = "Enable/Disable",
    width = "full",
    get = function() return self:GetConfig().enabled end,
    set = function(info, value) return self:UpdateConfig({enabled = value}) end,
    order = order or 1
  }
end

function visualPrototype:ToggleAnchorOption(order)
  return {
    type = "execute",
    name = "Show/Hide Anchor",
    func = function()
      self:ToggleAnchor()
    end,
    order = order or 1
  }
end

-----------------------------EXPORT--------------------------------------------

NayMods.plugins.visuals.prototype = visualPrototype
