-- ** NayMods.plugins.visuals.message **
-- A visual message on screen.
-------------------------------------------------------------------------------

local NayMods = NayMods
local Debug = NayMods.debug
local Utils = NayMods.utils
local VisualPrototype = NayMods.plugins.visuals.prototype
local Media = LibStub("LibSharedMedia-3.0")

local message = Utils:NewModule(VisualPrototype)

--[[
OnInit hook.
]]
function message:OnInit(key, options)
  if _G["NayMods_Message_FontString_" .. key] then
    --Debug:WriteLine("FontString with key %s already exists, re-using it.", key)
    self.fontString = _G["NayMods_Message_FontString_" .. key]
  else
    self.fontString = self.frame:CreateFontString("NayMods_Message_FontString_" .. key, "ARTWORK", "GameFontNormalLarge")
  end

  self:UpdateFont(self:GetConfig().font, self:GetConfig().fontSize)
  self.fontString:SetPoint("CENTER", self.frame, "CENTER", 0, 0)
  self.fontString:SetJustifyH(self:GetTextAlign())
end

--[[
OnShow hook.
]]
function message:OnShow()
  --Debug:WriteLine("OnShow")
  --Debug:Dump(self:GetConfig())
  self:UpdateFont(self:GetConfig().font, self:GetConfig().fontSize)
  self.frame:SetHeight(self.fontString:GetHeight())
  self.frame:SetWidth(self.fontString:GetWidth())
end

--[[
Updates the font and fontsize of the message.
]]
function message:UpdateFont(font, fontSize)
  --Debug:WriteLine("Update font = %s, size = %s", font, fontSize)

  if font ~= self:GetConfig().font then
    self:UpdateConfig({ font = font})
  end

  if fontSize ~= self:GetConfig().fontSize then
    self:UpdateConfig({ fontSize = fontSize })
  end

  local fontPath = Media:Fetch("font", font)
  self.fontString:SetFont(fontPath, fontSize)
  self.frame:SetHeight(self.fontString:GetHeight())
  self.frame:SetWidth(self.fontString:GetWidth())
end

function message:UpdateTextAlign(textAlign)
  if textAlign ~= self:GetConfig().textAlign then
    Debug:WriteLine("Updating align %s", textAlign)
    self:UpdateConfig({ textAlign = textAlign })
    self.fontString:SetJustifyH(textAlign)

    self.frame:SetHeight(self.fontString:GetHeight())
    self.frame:SetWidth(self.fontString:GetWidth())
  end
end

--[[
Gets the current font of the message.
]]
function message:GetFont()
  return self:GetConfig().font
end

--[[
Gets the current font size of the message.
]]
function message:GetFontSize()
  return self:GetConfig().fontSize
end

function message:GetTextAlign()
  return self:GetConfig().textAlign
end

--[[
Sets the text value of the message.
]]
function message:SetText(text)
  self.originalText = self.originalText or text
  self.text = text
  self.fontString:SetText(text)
  self.frame:SetHeight(self.fontString:GetHeight())
  self.frame:SetWidth(self.fontString:GetWidth())
end

--[[
Hook into the onUpdate of the visual prototype.
We enable dynamic text by doing so.
]]
message:RegisterOnUpdate("MSG_ENABLE_DYNAMIC_TEXT", function(...)
  local msg, totalElapsed = ...

  if not msg.duration then
    return
  end

  local timeLeft = msg.duration - totalElapsed
  local dynamicText = string.gsub(msg.originalText, "{D}", string.format("%.1f", timeLeft))
  msg:SetText(dynamicText)
end)

-----------------------------ACE3-OPTIONS-UTILS--------------------------------

function message:FontSelector(order)
  local this = self

  return {
    type = "select",
    dialogControl = "LSM30_Font",
    name = "Font",
    desc = "Font",
    values = AceGUIWidgetLSMlists.font,
    get = function()
      return this:GetFont()
    end,
    set = function(self, key)
      this:UpdateFont(key, this:GetFontSize())
    end,
    order = order or 1
  }
end

function message:FontSizer(order)
  return {
    type = "range",
    name = "Size",
    min = 8,
    max = 40,
    softMin = 8,
    softMax = 40,
    step = 1,
    bigStep = 4,
    get = function()
      return self:GetFontSize()
    end,
    set = function(info, value)
      self:UpdateFont(self:GetFont(), value)
    end,
    order = order or 1
  }
end

function message:TextAligner(order)
  local alignValues = {
    CENTER = "CENTER",
    LEFT = "LEFT",
    RIGHT = "RIGHT"
  }

  return {
    type = "select",
    name = "Text align",
    desc = "Text alignment",
    values = alignValues,
    get = function()
      return self:GetTextAlign()
    end,
    set = function(this, key)
      self:UpdateTextAlign(key)
    end,
    order = order or 1
  }
end

-----------------------------EXPORT--------------------------------------------

NayMods.plugins.visuals.message = message
